# PyBot is a python program for the purpose of making it possible to write darwinbots bot code in a more readable language similar to python (but with modifications to make it more useful for DarwinBots), and for making complex bots much more easily.
# Copyright (C) 2002-2003, 2006-2008 Richard D
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see http://www.gnu.org/licenses/.

# See PyBot.py for more information on the author.

from __future__ import division
import string

#Ready these patterns.

#Ready the patterns used for compiling

def addNormal(compiler):
	#Optypes:
	#	0 - Unary operator, e.g. !foo. foo is %1. In a script, this can be written "unary foo" or "unary(foo)" - both should be accepted
	#	-1 - Binary operator, e.g. foo + bar. foo is %1 and bar is %2. In a script, this can only be written "foo binary bar" - not "binary(foo, bar)"
	#	-2 - Direct operator - this has no arguments, and is used like a constant or variable name.
	#	A positive value - Function operator, where the value is the number of parameters. e.g. 3 for foo(bar,zort,narf). These must have two or more parameters.
	
	optype_unary = 0
	optype_binary = -1
	optype_direct = -2
	compiler.optype_unary = optype_unary
	compiler.optype_binary = optype_binary
	compiler.optype_direct = optype_direct
	
	compiler.conditionEnders = ("and", "or", "xor", ">", "<", ">=", "<=", "=", "!=", "%=", "!%=")	# These are things which should be at the end of a well-formed condition. These are read and checked to show a warning on obviously wrong conditions. (true and false are allowed as well, and 1 and 0 are translated to true and false, and are allowed)
	
	#Operators that can be translated:
	#Unary: inc, dec, rnd, sgn, abs, not, ~, ++, --, -, sqr, andbits (&), orbits (|), xorbits(^)
	#Binary: add, sub, mult, div, mod, and, or, xor, >, <, >=, <=, == (to =), !=, %=, !%=, >>, <<, store
	#Function: angle(2), dist(4?), ceil(2), floor(2), pow(2), pyth(2)
	#Direct: dup, true, false
	
	#Operators which aren't recognized by the translation routine:
	# store, inc, and dec are handled separately by writing foo = bar, foo++, or foo--. There ARE operators for them, but they are probably not useful.
	# *, which is handled by other pointer stuff ($ and &).
	# ~= and !~=, which are deprecated, according to the DarwinBots wiki.
	
	compiler.convert['normal']={}
	compiler.patterns['normal']={}
	addStandardOperator(compiler, 'normal', 'inc', optype_unary)
	addStandardOperator(compiler, 'normal', 'dec', optype_unary)
	addStandardOperator(compiler, 'normal', 'rnd', optype_unary)
	addStandardOperator(compiler, 'normal', 'rand', optype_unary, 'rnd')
	addStandardOperator(compiler, 'normal', 'sgn', optype_unary)
	addStandardOperator(compiler, 'normal', 'sign', optype_unary, 'sgn')
	addStandardOperator(compiler, 'normal', 'abs', optype_unary)
	addStandardOperator(compiler, 'normal', 'dup', optype_direct)
	addStandardOperator(compiler, 'normal', 'dupe', optype_direct, 'dup')
	addStandardOperator(compiler, 'normal', 'not', optype_unary)
	addStandardOperator(compiler, 'normal', '!', optype_unary, 'not')
	addStandardOperator(compiler, 'normal', '~', optype_unary)	#complement
	addStandardOperator(compiler, 'normal', '++', optype_unary)
	addStandardOperator(compiler, 'normal', '--', optype_unary)
	addStandardOperator(compiler, 'normal', '-u', optype_unary, '-')
	addStandardOperator(compiler, 'normal', 'sqr', optype_unary)
	addStandardOperator(compiler, 'normal', 'sqrt', optype_unary, 'sqr')
	addStandardOperator(compiler, 'normal', 'andbits', optype_binary, '&')
	addStandardOperator(compiler, 'normal', 'orbits', optype_binary, '|')
	addStandardOperator(compiler, 'normal', 'xorbits', optype_binary, '^')
	
	addStandardOperator(compiler, 'normal', '+', optype_binary, 'add')
	addStandardOperator(compiler, 'normal', '-b', optype_binary, 'sub')
	addStandardOperator(compiler, 'normal', '*', optype_binary, 'mult')
	addStandardOperator(compiler, 'normal', '/', optype_binary, 'div')
	addStandardOperator(compiler, 'normal', '%', optype_binary, 'mod')
	addStandardOperator(compiler, 'normal', 'and', optype_binary)
	addStandardOperator(compiler, 'normal', 'or', optype_binary)
	addStandardOperator(compiler, 'normal', 'xor', optype_binary)
	addStandardOperator(compiler, 'normal', '>', optype_binary)
	addStandardOperator(compiler, 'normal', '<', optype_binary)
	addStandardOperator(compiler, 'normal', '>=', optype_binary)
	addStandardOperator(compiler, 'normal', '<=', optype_binary)
	addStandardOperator(compiler, 'normal', '==', optype_binary, '=')
	addStandardOperator(compiler, 'normal', '!=', optype_binary)
	addStandardOperator(compiler, 'normal', '%=', optype_binary)
	addStandardOperator(compiler, 'normal', '!%=', optype_binary)
	addStandardOperator(compiler, 'normal', '>>', optype_binary)
	addStandardOperator(compiler, 'normal', '<<', optype_binary)
	addStandardOperator(compiler, 'normal', 'store', optype_binary)
	
	addStandardOperator(compiler, 'normal', 'angle', 2)
	addStandardOperator(compiler, 'normal', 'dist', 2)
	addStandardOperator(compiler, 'normal', 'max', 2, 'ceil')
	addStandardOperator(compiler, 'normal', 'min', 2, 'floor')
	addStandardOperator(compiler, 'normal', 'ceil', 2)
	addStandardOperator(compiler, 'normal', 'floor', 2)
	addStandardOperator(compiler, 'normal', 'pow', 2)
	addStandardOperator(compiler, 'normal', 'pyth', 2)
	
	# These were, IIRC, an experiment, and aren't documented in the manual because they aren't useful at this time. They would be useful if memval was updated instantly after memloc was set - then you could write 'memat(42)' and that would set .memloc to 42 and return *.memval.
	addStandardOperator(compiler, 'normal', 'memat', optype_unary, '.memloc store *.memval')
	addStandardOperator(compiler, 'normal', 'tmemat', optype_unary, '.tmemloc store *.tmemval')
	
	addStandardOperator(compiler, 'normal', 'true', optype_direct)
	addStandardOperator(compiler, 'normal', 'false', optype_direct)
	
	#-----------------------------------------
	compiler.convert['sg']={}
	compiler.patterns['sg']={}
	addStandardOperator(compiler, 'sg', 'inc', optype_unary)
	addStandardOperator(compiler, 'sg', 'dec', optype_unary)
	addStandardOperator(compiler, 'sg', 'rnd', optype_unary)
	addStandardOperator(compiler, 'sg', 'rand', optype_unary, 'rnd')
	addStandardOperator(compiler, 'sg', 'sgn', optype_unary)
	addStandardOperator(compiler, 'sg', 'sign', optype_unary, 'sgn')
	addStandardOperator(compiler, 'sg', 'abs', optype_unary)
	addStandardOperator(compiler, 'sg', 'dup', optype_direct)
	addStandardOperator(compiler, 'sg', 'dupe', optype_direct, 'dup')
	addStandardOperator(compiler, 'sg', 'not', optype_unary, '- ++')
	addStandardOperator(compiler, 'sg', '!', optype_unary, '- ++')
	addStandardOperator(compiler, 'sg', '~', optype_unary)
	addStandardOperator(compiler, 'sg', '++', optype_unary)
	addStandardOperator(compiler, 'sg', '--', optype_unary)
	addStandardOperator(compiler, 'sg', '-u', optype_unary, '-')
	addStandardOperator(compiler, 'sg', 'sqr', optype_unary)
	addStandardOperator(compiler, 'sg', 'sqrt', optype_unary, 'sqr')
	addStandardOperator(compiler, 'sg', 'andbits', optype_binary, '&')
	addStandardOperator(compiler, 'sg', 'orbits', optype_binary, '|')
	addStandardOperator(compiler, 'sg', 'xorbits', optype_binary, '^')
	
	addStandardOperator(compiler, 'sg', '+', optype_binary, 'add')
	addStandardOperator(compiler, 'sg', '-b', optype_binary, 'sub')
	addStandardOperator(compiler, 'sg', '*', optype_binary, 'mult')
	addStandardOperator(compiler, 'sg', '/', optype_binary, 'div')
	addStandardOperator(compiler, 'sg', '%', optype_binary, 'mod')
	addStandardOperator(compiler, 'sg', 'and', optype_binary, '&')
	addStandardOperator(compiler, 'sg', 'or', optype_binary, '|')
	addStandardOperator(compiler, 'sg', 'xor', optype_binary, '^')
	
	addStandardOperator(compiler, 'sg', '>', optype_binary, 'sub sgn 0 floor')
	addStandardOperator(compiler, 'sg', '<', optype_binary, 'sub sgn - 0 floor')
	addStandardOperator(compiler, 'sg', '>=', optype_binary, 'sub 1 add sgn 0 floor')
	addStandardOperator(compiler, 'sg', '<=', optype_binary, 'sub 1 sub sgn - 0 floor')
	addStandardOperator(compiler, 'sg', '==', optype_binary, 'sub sgn abs - ++')
	addStandardOperator(compiler, 'sg', '!=', optype_binary, 'sub sgn abs')
	addStandardOperator(compiler, 'sg', '%=', optype_binary)
	addStandardOperator(compiler, 'sg', '!%=', optype_binary)
	addStandardOperator(compiler, 'sg', '>>', optype_binary)
	addStandardOperator(compiler, 'sg', '<<', optype_binary)
	addStandardOperator(compiler, 'sg', 'store', optype_binary)
	
	addStandardOperator(compiler, 'sg', 'angle', 2)
	addStandardOperator(compiler, 'sg', 'dist', 2)
	addStandardOperator(compiler, 'sg', 'max', 2, 'ceil')
	addStandardOperator(compiler, 'sg', 'min', 2, 'floor')
	addStandardOperator(compiler, 'sg', 'ceil', 2)
	addStandardOperator(compiler, 'sg', 'floor', 2)
	addStandardOperator(compiler, 'sg', 'pow', 2)
	addStandardOperator(compiler, 'sg', 'pyth', 2)
	
	# There's a comment about memat and tmemat in the "normal" section.
	addStandardOperator(compiler, 'sg', 'memat', optype_unary, '.memloc store *.memval')
	addStandardOperator(compiler, 'sg', 'tmemat', optype_unary, '.tmemloc store *.tmemval')
	
	addStandardOperator(compiler, 'sg', 'true', optype_direct, '1')
	addStandardOperator(compiler, 'sg', 'false', optype_direct, '0')
	

def addStandardOperator(compiler, patternName, opname, optype, translation=None):
	compiler.patterns[patternName][opname]={}
	compiler.patterns[patternName][opname]['optype']=optype
	compiler.patterns[patternName][opname]['standard']=1
	if (translation==None):
		translation = opname
	compiler.patterns[patternName][opname]['translation']=translation

	