# PyBot (for DarwinBots) is a python program for the purpose of making it possible to write darwinbots bot code in a more readable language similar to python (but with modifications to make it more useful for DarwinBots), and for making complex bots much more easily.
# Copyright (C) 2002-2003, 2006-2008 Richard D
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see http://www.gnu.org/licenses/.

# The author is, on the DarwinBots forum, known as Trafalgar.
# PyBot was originally written in 2007, and completed and released in 2008.
# I based PyBot on another python program I had written which did something similar for a different language - that program was called PySphere and was written in 2002-2003 and 2006. I ripped out a lot of the code but kept the useful framework and useful functions and such.


releaseNumber='2'

enableTimeChecking=1

import os, sys, os.path, string
import cPickle
import getpass
import telnetlib
import time
import socket
import Patterns

#First thing that's run. Well, second thing. The very last line of this file is the first thing, and it calls this.
def main():
	print "PyBot V"+releaseNumber+" - Copyright 2002-2003 and 2006-2008 Richard D"
	os.chdir(os.path.join('..','Robots'))
	compiler = Compiler()
	
	compilePyr(os.listdir("."), compiler)
	
	renamePyrc(os.listdir("."), compiler)
	
	print "Done!"

#This takes a list of files, and compiles all .pyr's in it, and recurses into all folders.
def compilePyr(files, compiler):
	for file in files:
		if file[-4:]=='.pyr':
			compiler.compileFile(file)
		elif os.path.isdir(file):
			#print "Searching "+file
			os.chdir(file)
			compilePyr(os.listdir("."), compiler)
			os.chdir("..")

def renamePyrc(files, compiler):
	for file in files:
		if file[-5:]=='.pyrc':
			compiler.renameFile(file)
		elif os.path.isdir(file):
			#print "Searching "+file
			os.chdir(file)
			renamePyrc(os.listdir("."), compiler)
			os.chdir("..")
			
#The compiler class. I use self. stuff a lot with this.
class Compiler:
	def __init__(self):
		self.patterns={}
		self.convert={}
		self.defined={}
		self.scp=None
		self.numTabs=0
		self.lineTabs=0
		Patterns.addNormal(self)
	
	#Compile one particular file.
	def compileFile(self, file):
		#print "Compile "+file
		pycFile = file[:-4]+'.pyrc'
		scpFile = file[:-4]+'.txt'
		pyt=os.stat(file).st_mtime
		if os.path.exists(scpFile):
			scpt=os.stat(scpFile).st_mtime
		else:
			scpt=0
		#print file+"'s time is "+str(pyt)+" and "+scpFile+"'s time is "+str(scpt)
		if (enableTimeChecking and pyt<scpt):
			return
		self.moduleName = file[:-4]
		print "Compiling "+file+" to "+pycFile
		self.py=open(file,'r')
		self.scp=open(pycFile,'w')
		self.scp.write('\''+scpFile+' compiled by PyBot V'+releaseNumber+' from '+file+'.\n')
		self.enableStorecheck=1
		self.includeOriginalCode=1
		self.sgMore=0
		self.combineLines=0
		self.inlineDefs=0
		self.extraText=[]
		self.debug=0
		self.numDefsWritten=0
		self.writeToExtraText=0
		self.inScriptDef=0
		self.scriptVer='normal'
		eof=0
		self.lineNum=0
		self.defined={}
		self.blocks=[]
		self.extraTextBlocks=[]
		self.blockTab=0
		self.firstComments=1
		self.inConditions=0
		self.branchUse=0
		self.branchValueSet=0
		
		while eof==0:
			line = self.py.readline()
			self.lineNum+=1
			if line[-1:]=='\n':
				line=line[:-1]
			else:
				eof=1
			self.parseline(line)
		self.parseline('\t')
		while self.blockTab>0:
			self.lineTabs=self.blockTab-1
			st = self.blocks.pop()
			if st!=None:
				if st.__class__==str:
					self.scp_write(st)
				else:
					if st==1:		#elif - pop three
						output = 'add add dup ^ inc\n'
						self.inConditions-=3
						self.branchValueSet=0
					elif st==3:		#if inside another if - pop 2
						output = 'add dup ^ inc\n'
						self.inConditions-=2
						self.branchValueSet=0
					else:	#if or else - pop one
						output='dup ^ inc\n'
						self.inConditions-=1
						self.branchValueSet=0
					self.scp_write(output)
			st = self.extraTextBlocks.pop()
			if st[0]!=None:
				self.extraText.append(st[0])
			if st[1]==1:
				self.writeToExtraText=0
			self.blockTab-=1
		for t in self.extraText:
			self.scp.write(t)
		self.py.close()
		self.scp.close()
		if (self.numDefsWritten>50):
			print "Warning: DarwinBots 2.43 crashes if you have more than 50 defs in a bot. You have "+str(self.numDefsWritten)+" in "+scpFile+". If DarwinBots crashes when you try to start a match with the bot, this would probably be why."
		
	def renameFile(self, file):
		pycFile = file[:-5]+'.pyrc'
		scpFile = file[:-5]+'.txt'
		print "Renaming "+pycFile+" to "+scpFile+"."
		if os.path.exists(scpFile):
			os.remove(scpFile)
		os.rename(pycFile,scpFile)
	
	
	def scp_write(self,arg):
		if arg.lower()=='pass\n':
			return
		arg=arg.strip(' \t')
		tabs = self.lineTabs
		#if (tabs>1):
		#	arg=''*(tabs-1)+arg
		if self.writeToExtraText==1:
			self.extraText.append(arg)
		else:
			self.scp.write(arg)
			
	#Parse one line of a file.
	def parseline(self,line):
		self.defnameParse=1
		strline = line
		commentPos = self.finds(strline,'#',0)
		if commentPos>-1:
			if strline.strip().lower()=="#include-original-code":
				self.includeOriginalCode=1
			elif strline.strip().lower()=="#omit-original-code":
				self.includeOriginalCode=0
			elif strline.strip().lower()=="#combine-lines":
				self.combineLines=1
			elif strline.strip().lower()=="#inline-defs":
				self.inlineDefs=1
			elif strline.strip().lower()=="#sg-more":
				self.sgMore=1
			elif strline.strip().lower()=="#sg-normal":
				self.sgMore=0
			if len(strline.strip())>8 and strline.strip()[:8].lower()=="#script-":
				self.scriptVer=strline.strip()[8:]
			if len(strline.strip())>7 and strline.strip()[:7].lower()=="#debug-":
				self.debug=int(strline.strip()[7:])
			if strline.strip().lower()=="#enable-storecheck":
				self.enableStorecheck=1
			elif strline.strip().lower()=="#disable-storecheck":
				self.enableStorecheck=0
			if len(strline.strip())>12 and strline.strip()[:12].lower()=="#branch-use ":
				self.branchUse=strline.strip()[12:].strip(' \t')
				#if self.debug:
				#	print "branch-use prior to comment stripping is "+self.branchUse
				commentPos2 = self.finds(self.branchUse,'#',0)
				if commentPos2>-1:
					self.branchUse=self.branchUse[:commentPos2].strip(' \t')
				#if self.debug:
				#	print "branch-use after comment stripping is "+self.branchUse
				#Disabled 5/8/2008: val = '0 *'+self.branchUse+' 1 sub sgn 0 floor *'+self.branchUse+' sgn - 0 floor | '+self.branchUse+' mult store\n'
				#X-1 sgn: -1 means x<=0. 1 means x>=2. 0 means x==1. floor zero means x>1.
				#-sgn(X): -1 means x>=1. 1 means x<=-1. 0 means x==0. floor zero means x<0.

				#print 'Writing ('+val+')'
				#Disabled 5/8/2008: self.scp_write(val)
			
				
			strline=strline[:commentPos]
		if self.includeOriginalCode:
			if (len(strline)!=0 and self.firstComments==1):
				self.scp.write('\n')
				self.firstComments=0
			if (len(line.strip(' \t#'))>0):
				self.scp_write('\''+line.strip('#')+'\n')	# - Conditions '+str(self.inConditions)+
		if len(strline)==0:
			return ""
		while strline[-1:]=='\t' or strline[-1:]==' ':
			strline=strline[:-1]
		if not hasContent(strline):
			return ""
		numTabs = self.countTabs(strline)
		if len(strline[numTabs:])==0:
			return ""
		self.numTabs=numTabs
		self.lineTabs=numTabs
		
		if len(strline)>4:
			if strline[:4].lower()=='def ':
				#Variable name declaration, just strip out the ' at ' and pass it through
				#Example: def foo at 42						# def foo 42
				self.inScriptDef=0
				#~ while numTabs<self.blockTab:
					#~ self.lineTabs=self.blockTab-1
					#~ st = self.blocks.pop()
					#~ if st!=None:
						#~ self.scp_write(st)
					#~ st = self.extraTextBlocks.pop()
					#~ if st[0]!=None:
						#~ self.extraText.append(st[0])
					#~ if st[1]==1:
						#~ self.writeToExtraText=0
					#~ self.blockTab-=1
				
				atpos = strline.find(' at ')
				if atpos==-1:
					raise SystemExit,"Missing 'at' in def (Expected something like 'def foo at 42'): Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				if self.inlineDefs:
					defineName = strline[4:atpos].strip()
					defineVal = strline[atpos+4:].strip()
					self.defined[defineName] = (defineVal, 1)
				else:
					self.numDefsWritten+=1
					self.scp_write(strline[0:atpos]+' '+strline[atpos+4:]+'\n')
				return
			elif strline[:3].lower()=='if ':
				self.branchValueSet=0
				#cond-start
				ifpos = 0
				self.scriptVer='normal'
				while numTabs<self.blockTab:
					self.lineTabs=self.blockTab-1
					st = self.blocks.pop()
					if st.__class__==str:
						self.scp_write(st)
					else:
						if st==1:		#elif - pop three
							output = 'dropbool dropbool dropbool\n' #'add add dup ^ inc\n'
							self.inConditions-=3
						elif st==3:		#if inside another if - pop 2
							output = 'dropbool dropbool\n' #'add dup ^ inc\n'
							self.inConditions-=2
						else:	#if or else - pop one
							output= 'dropbool\n' #'dup ^ inc\n'
							self.inConditions-=1
						self.scp_write(output)
						
					st = self.extraTextBlocks.pop()
					if st[0]!=None:
						self.extraText.append(st[0])
					if st[1]==1:
						self.writeToExtraText=0
					self.blockTab-=1
				colon = strline.find(':')
				if colon==-1:
					raise SystemExit,"Missing colon at the end of if: Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				condition=self.conversionMagic(strline[ifpos+3:colon], startPointers=1)
				if condition==None:
					condition=""
				if condition.__class__!=str:
					condition=str(condition)
				backn = '\n'
				if self.combineLines:
					backn=' '
				self.scp_write("cond"+backn)
				self.scp_write("\t"+condition+backn)
				self.scp_write("start"+backn)
				self.blocks.append('stop'+backn+backn)
				self.extraTextBlocks.append(('',0))
				self.blockTab+=1
				self.scriptVer='sg'
				return
			elif strline[:4].lower()=='else':
				self.branchValueSet=0
				#else
				elsepos = 0
				numTabs+=1
				while numTabs<self.blockTab:
					self.lineTabs=self.blockTab-1
					st = self.blocks.pop()
					if st.__class__==str:
						self.scp_write(st)
					else:
						if st==1:		#elif - pop three
							output = 'dropbool dropbool dropbool\n' #'add add dup ^ inc\n'
							self.inConditions-=3	#bugfix: was 2
						elif st==3:
							output = 'dropbool dropbool\n' #'add dup ^ inc\n'
							self.inConditions-=2
						else:	#if or else - pop one
							output = 'dropbool\n' #'dup ^ inc\n'
							self.inConditions-=1
						self.scp_write(output)
						#self.scp_write("inConditions is now "+str(self.inConditions))
						
					st = self.extraTextBlocks.pop()
					if st[0]!=None:
						self.extraText.append(st[0])
					if st[1]==1:
						self.writeToExtraText=0
					self.blockTab-=1
				colon = strline.find(':')
				if colon==-1:
					raise SystemExit,"Missing colon at the end of else: Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				if (len(strline[elsepos+4:colon].strip())>0):
					raise SystemExit,"Cannot have a condition in an 'else': Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				backn = '\n'
				if self.combineLines:
					backn=' '
				self.scp_write("else"+backn)
				self.scriptVer='sg'
				return
			elif (strline.strip(' \t').find('define ')==0):
				colon = strline.find(':')
				if colon==-1:
					raise SystemExit,'Missing : in define statement on line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				else:
					define = strline.find('define ')
					defineName = strline[define+7:colon].strip()
					defineVal = strline[colon+1:].strip()
					self.defined[defineName] = (defineVal, 0)
		if len(strline)>1:
			if strline[:1]=='\t':
				#special cases
				if self.debug:
					print "numTabs on line "+str(self.lineNum)+" is "+str(numTabs)+". blockTab is "+str(self.blockTab)
				if numTabs>self.blockTab:
					raise SystemExit,'Unexpected indentation on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				strline=strline[numTabs:]
				
				if 1:
					lastBlockStatement = None
					while numTabs<self.blockTab:
						self.lineTabs=self.blockTab-1
						st = self.blocks.pop()
						if st.__class__==str:
							self.scp_write(st)
						else:
							if (numTabs == self.blockTab-1):
								lastBlockStatement = st
							else:
								if st==1:		#elif - pop three
									output = 'dropbool dropbool dropbool\n' #'add add dup ^ inc\n'
									self.inConditions-=3
									self.branchValueSet=0
								elif st==3:
									output = 'dropbool dropbool\n' #'add dup ^ inc\n'
									self.inConditions-=2
									self.branchValueSet=0
								else:	#if or else - pop one
									output = 'dropbool\n' #'dup ^ inc\n'
									self.inConditions-=1
									self.branchValueSet=0
								self.scp_write(output)
						st = self.extraTextBlocks.pop()
						if st[0]!=None:
							self.extraText.append(st[0])
						if st[1]==1:
							self.writeToExtraText=0
						self.blockTab-=1
					when = strline.find(' when ')
					equals = strline.find('=')
					equalsequals = strline.find('==')
					notequals = strline.find('!=')
					gtequals = strline.find('>=')
					ltequals = strline.find('<=')
					pequals = strline.find('%=')
					cequals = strline.find('~=')
					inc = strline.find('++')
					dec = strline.find('--')
					define = strline.find('define ')
					ifpos = strline.find('if ')
					elsepos = strline.find('else')
					elifpos = strline.find('elif ')
					if (strline[-1]==':') and ifpos==-1 and elsepos==-1 and elifpos==-1:
						raise SystemExit,'We don\'t know what the block keyword in ('+strline+') is, on line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						return
					if (define!=-1):
						if (strline.strip(' \t').find('define')==0):
							colon = strline.find(':')
							if colon==-1:
								raise SystemExit,'Missing : in define statement on line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
							else:
								defineName = strline[define+7:colon].strip()
								defineVal = strline[colon+1:].strip()
								self.defined[defineName] = (defineVal, 0)
					elif (elifpos!=-1):
						#if (self.branchUse==0):
						#	raise SystemExit,'You need to write #branch-use 700 (or some other number) on a line before you can use branch conditionals. Line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						self.branchValueSet=0
						if self.debug:
							print "elif on line "+str(self.lineNum)+"."
						beforeLine = ''
						afterLine = ''
						if lastBlockStatement==0:		#if
							beforeLine='dupbool not dupbool ' #'dup - ++ dup '
							afterLine=' and' #' &'
							self.inConditions+=2
						elif lastBlockStatement==1:		#elif
							beforeLine='not and dupbool ' #'- ++ & dup '
							afterLine=' and' #' &'
							#self.inConditions doesn't change
						elif lastBlockStatement==2:		#else
							#error
							raise SystemExit,"Error: You can't have an elif following an else. Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						elif lastBlockStatement==3:		#if inside another if
							#stack has (outside if junk ending in a T/F), then a copy of that, then T/F for whether this if was run.
							#XTT do not run. XFF do not run. XTF run.
							#- ++ changes it to XTF do not run, XFT do not run, XTT run.
							#& changes it to XF do not run or XT run.
							#- ++ changes it to XT do not run or XF run, which is what it normally would have been if this were an outermost if.
							
							# dup *.bar 3 sub sgn abs - ++ & (stack holds 2 values now)
							## dupbool *.bar 3 = and (boolean stack holds 2 values now)
							beforeLine = 'not and not dupbool not dupbool ' #'- ++ & - ++ dup - ++ dup '
							afterLine=' and' #' &'
							self.inConditions+=1
						else:
							#error
							raise SystemExit,"Error: Elif without if statement before it. Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						colon = strline.find(':')
						if colon==-1 or colon<elifpos:
							raise SystemExit,"Missing colon at the end of elif: Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						self.scriptVer='normal'
						condition=self.conversionMagic(strline[elifpos+5:colon], startPointers=1)
						self.scriptVer='sg'
						if condition==None:
							condition="true"
						if condition.__class__!=str:
							condition=str(condition)
						if (condition.lower() == "1"):
							condition="true";
						elif (condition.lower() == "0"):
							condition="false";
						elif ((condition.lower()!="true") and (condition.lower()!="false")):
							if (not condition.endswith(self.conditionEnders)):
								print self.conditionWarningMessage(strline[elifpos+5:colon].strip(), condition)
						backn = '\n'
						if self.combineLines:
							backn=' '
						if self.debug:
							print "BL("+beforeLine+") AL("+afterLine+")"
						self.scp_write(beforeLine+condition+afterLine+backn)
						self.blocks.append(1)
						self.extraTextBlocks.append(('',0))
						self.blockTab+=1
						numTabs+=1
						if self.debug:
							print "blockTab now "+str(self.blockTab)+", inConditions now "+str(self.inConditions)
					elif (ifpos!=-1):
						#if (self.branchUse==0):
						#	raise SystemExit,'You need to write #branch-use 700 (or some other number) on a line before you can use branch conditionals. Line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						self.branchValueSet=0
						if self.debug:
							print "if on line "+str(self.lineNum)+"."
						beforeLine = ''
						afterLine = ''
						if lastBlockStatement==0:		#if - pop one
							beforeLine = 'dropbool ' #'dup ^ inc '
							self.inConditions-=1
						elif lastBlockStatement==1:		#elif - pop three
							beforeLine = 'dropbool dropbool dropbool ' #'add add dup ^ inc '
							self.inConditions-=3
						elif lastBlockStatement==2:		#else - pop one
							beforeLine = 'dropbool ' #'dup ^ inc '
							self.inConditions-=1
						elif lastBlockStatement==3:		#if inside another if
							beforeLine = 'dropbool dropbool ' #'add dup ^ inc'	#pop two
							self.inConditions-=2
							
						if self.inConditions>0:
							beforeLine=beforeLine+'dupbool dupbool ' #'dup dup '
							afterLine=' and' #' &'
						colon = strline.find(':')
						if colon==-1 or colon<ifpos:
							raise SystemExit,"Missing colon at the end of if: Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						self.scriptVer='normal'
						condition=self.conversionMagic(strline[ifpos+3:colon], startPointers=1)
						self.scriptVer='sg'
						if condition==None:
							condition="true"
						if condition.__class__!=str:
							condition=str(condition)
						if (condition.lower() == "1"):
							condition="true";
						elif (condition.lower() == "0"):
							condition="false";
						elif ((condition.lower()!="true") and (condition.lower()!="false")):
							if (not condition.endswith(self.conditionEnders)):
								print self.conditionWarningMessage(strline[ifpos+3:colon].strip(), condition)
						backn = '\n'
						if self.combineLines:
							backn=' '
						if self.debug:
							print "BL("+beforeLine+") AL("+afterLine+")"
						self.scp_write(beforeLine+condition+afterLine+backn)
						if self.inConditions>0:
							self.blocks.append(3)
							self.inConditions+=1
						else:
							self.blocks.append(0)
						self.extraTextBlocks.append(('',0))
						self.blockTab+=1
						self.inConditions+=1
						numTabs+=1
						if self.debug:
							print "blockTab now "+str(self.blockTab)+", inConditions now "+str(self.inConditions)
					elif (elsepos!=-1):
						#if (self.branchUse==0):
						#	raise SystemExit,'You need to write #branch-use 700 (or some other number) on a line before you can use branch conditionals. Line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						self.branchValueSet=0
						if self.debug:
							print "else on line "+str(self.lineNum)+"."
						beforeLine = ''
						if lastBlockStatement==0:		#if
							beforeLine='not ' #'- ++ '
						elif lastBlockStatement==1:		#elif
							beforeLine='not and not or not ' #'- ++ & - ++ | - ++ '
							self.inConditions-=2
						elif lastBlockStatement==3:		#if inside another if
							#stack has (outside if junk ending in a T/F), then a copy of that, then T/F for whether this if was run.
							#XTT do not run. XFF do not run. XTF run.
							#- ++ changes it to XTF do not run, XFT do not run, XTT run.
							#& changes it to XF do not run or XT run.
							beforeLine='not and ' #'- ++ & '
							self.inConditions-=1
						elif lastBlockStatement==2:		#else
							#error
							raise SystemExit,"Error: You can't have an else following another else. Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						else:
							#error
							raise SystemExit,"Error: else without if or elif statement before it. Line "+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						backn = '\n'
						if self.combineLines:
							backn=' '
						if self.debug:
							print "BL("+beforeLine+")"
						self.scp_write(beforeLine+backn)
						self.blocks.append(2)
						self.extraTextBlocks.append(('',0))
						self.blockTab+=1
						numTabs+=1
						if self.debug:
							print "blockTab now "+str(self.blockTab)+", inConditions now "+str(self.inConditions)
					else:
						if lastBlockStatement!=None:
							if lastBlockStatement==1:		#elif - pop three
								output = 'dropbool dropbool dropbool\n' #'add add dup ^ inc\n'
								self.inConditions-=3
							elif lastBlockStatement==3:
								output='dropbool dropbool\n' #'add dup ^ inc\n'
								self.inConditions-=2
							else:	#if or else - pop one
								output='dropbool\n' #'dup ^ inc\n'
								self.inConditions-=1
							self.branchValueSet=0
							self.scp_write(output)
						if (self.branchValueSet==0) and (self.inConditions>0) :
							#dup *700 - ++ & 700 mult inc dup - ++ *700 & 700 mult dec
							#Disabled 5/8/2008: self.scp_write('dup *'+self.branchUse+' - ++ & '+self.branchUse+' mult inc dup - ++ *'+self.branchUse+' & '+self.branchUse+' mult dec\n')
							self.branchValueSet=1
						beforeLine = ''
						afterAddress = ''
						afterStore = ''
						#if self.inConditions>0:
							#No longer needed 5/8/2008: afterAddress = ' *'+self.branchUse+' mult'
							#3694 dup *.out1 sub sgn abs .out1 mult *700 mult store
						while (equals!=-1 and (equals==equalsequals or equals==notequals+1 or equals==gtequals+1 or equals==ltequals+1 or equals==pequals+1 or equals==cequals+1)):
							nextpos = equals+2
							equals = strline.find('=', nextpos)
							equalsequals = strline.find('==', nextpos)
							notequals = strline.find('!=', nextpos)
							gtequals = strline.find('>=', nextpos)
							ltequals = strline.find('<=', nextpos)
							pequals = strline.find('%=', nextpos)
							cequals = strline.find('~=', nextpos)
						if equals>-1 and not self.inquote(strline,equals):
							#variable assignment
							if equals==0:
								raise SystemExit,'WTF is line '+str(self.lineNum)+' of '+self.moduleName+'.pyr supposed to be?'
							else:
								# Support for +=, -=, etc
								tfndstr='+-*/%'
								tfnd=string.find(tfndstr,strline[equals-1])
								if tfnd>-1:
									strline = strline[0:equals-1]+'='+strline[0:equals-1]+strline[equals-1]+strline[equals:]
									equals -= 1
								
								tt=self.conversionMagic(strline[equals+1:].strip(), startPointers=1)
								if tt.__class__!=str:
									tt=str(tt)
								self.defined['alreadyset']=("("+strline[equals+1:]+")", 0)
								if (tt==None):
									raise SystemExit,'Error: Nothing recognized after equals sign on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
								sss=''
								if (when>-1):
									varname = self.conversionMagic(strline[:when].strip(), startPointers=0, maxTokens=1)
									if varname.__class__!=str:
										varname=str(varname)
									
									if self.sgMore:
										self.scriptVer = 'sg'
									else:
										self.scriptVer = 'normal'
									condition = self.conversionMagic(strline[when+6:equals].strip(), startPointers=1)
									self.scriptVer='sg'
									if condition.__class__!=str:
										condition=str(condition)
									if (self.sgMore):
										if (condition.lower() == "true"):
											condition="1";
										elif (condition.lower() == "false"):
											condition="0";
										sss = varname+" "+condition+" mult"
										
									else:
										if (condition.lower() == "1"):
											condition="true";
										elif (condition.lower() == "0"):
											condition="false";
										elif ((condition.lower()!="true") and (condition.lower()!="false")):
											if (not condition.endswith(self.conditionEnders)):
												print self.conditionWarningMessage(strline[when+6:equals].strip(), condition)
										if (self.inConditions>0):
											sss = varname+" dupbool "+condition+" and" #removed " mult" 5/8/2008
										else:
											sss = varname+" "+condition #removed " mult" 5/8/2008
									
									if (not self.sgMore):
										afterStore = " dropbool"
									if (tfnd==-1):
										if (self.enableStorecheck):
											#sss = "dup "+self.conversionMagic(strline[:when].strip(), startPointers=1, maxTokens=1)+" sub sgn abs "+condition+" & "+varname+" mult"
											if self.sgMore:
												self.scriptVer = 'sg'
											else:
												self.scriptVer = 'normal'
											if (self.sgMore):
												sss = "dup "+self.conversionMagic(strline[:when].strip(), startPointers=1, maxTokens=1)+" sub sgn abs "+condition+" & "+varname+" mult"
											else:
												sss = "dup "+self.conversionMagic(strline[:when].strip(), startPointers=1, maxTokens=1)+" != "+condition+" and "
											self.scriptVer='sg'
											if (self.inConditions>0):
												sss = "dupbool "+sss+"and "+varname
											else:
												sss = sss+varname
									
								else:
									sss=self.conversionMagic(strline[:equals].strip(), startPointers=0, maxTokens=1)	#maxTokens should cause an error to be thrown if more than one token is found.
								if (sss==None):
									raise SystemExit,'Error: Nothing recognized before equals sign on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
								if sss.__class__!=str:
									sss=str(sss)
								self.defined['alreadyset']=('0', 0)
								if (tfnd==-1 and when==-1):
									if (self.enableStorecheck):
										if self.sgMore:
											self.scriptVer = 'sg'
										else:
											self.scriptVer = 'normal'
										if (self.sgMore):
											if (sss.find(" mult")==-1):
												presss="dup "+self.conversionMagic(strline[:equals].strip(), startPointers=1, maxTokens=1)+" sub sgn abs "+sss+" mult"
										else:
											presss="dup "+self.conversionMagic(strline[:equals].strip(), startPointers=1, maxTokens=1)+" != "
										self.scriptVer='sg'
										if ((self.inConditions>0) and (not self.sgMore)):
											sss = "dupbool "+presss+"and "+sss
											afterStore = " dropbool"
										else:
											sss = presss+sss
										
										#'	eyeComp = (refeye==myeye)
										#*.refeye *.myeye sub sgn abs - ++ .eyeComp store
										#'	(eyeComp * ($eyeComp!=($refeye==$myeye))) = (refeye==myeye)
										#*.refeye *.myeye sub sgn abs - ++ .eyeComp *.eyeComp *.refeye *.myeye sub sgn abs - ++ sub sgn abs mult store

								backn = '\n'
								if self.combineLines:
									backn=' '
								self.scp_write(beforeLine+tt+" "+sss+afterAddress+" store"+afterStore+backn)
						elif inc>-1 and not self.inquote(strline,equals):
							incText = ''
							if (inc==0):
								incText = strline[inc+2:]
							elif (inc==len(strline)-2):
								incText = strline[:-2]
							else:
								raise SystemExit,'++ at unexpected position in line '+str(self.lineNum)+' of '+self.moduleName+'.pyr. ++ without an = is treated as inc, but must either be at the beginning or the end of the line.'
							when = incText.find(' when ')
							if (when>-1):
								varname = self.conversionMagic(incText[:when].strip(), startPointers=0, maxTokens=1)
								if varname.__class__!=str:
									varname=str(varname)
							
								if self.sgMore:
									self.scriptVer = 'sg'
								else:
									self.scriptVer = 'normal'
								condition = self.conversionMagic(incText[when+6:].strip(), startPointers=1)
								self.scriptVer='sg'
								if condition.__class__!=str:
									condition=str(condition)
								if (self.sgMore):
									if (condition.lower() == "true"):
										condition="1";
									elif (condition.lower() == "false"):
										condition="0";
									sss = sss = varname+" "+condition+" mult"
								else:
									if (condition.lower() == "1"):
										condition="true";
									elif (condition.lower() == "0"):
										condition="false";
									elif ((condition.lower()!="true") and (condition.lower()!="false")):
										if (not condition.endswith(self.conditionEnders)):
											print self.conditionWarningMessage(strline[when+6:].strip(), condition)
									if (self.inConditions>0):
										sss = varname+" dupbool "+condition+" and" #removed " mult" 5/8/2008
									else:
										sss = varname+" "+condition #removed " mult" 5/8/2008
								
								if (not self.sgMore):
									afterStore = " dropbool"
								
								
							else:
								sss=self.conversionMagic(incText, startPointers=0, maxTokens=1)	#maxTokens should cause an error to be thrown if more than one token is found.
							
							if (sss==None):
								raise SystemExit,'Error: Nothing recognized after ++ on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
							if sss.__class__!=str:
								sss=str(sss)
							backn = '\n'
							if self.combineLines:
								backn=' '
							self.scp_write(beforeLine+sss+afterAddress+" inc"+afterStore+backn)
						elif dec>-1 and not self.inquote(strline,equals):
							decText = ''
							if (dec==0):
								decText = strline[dec+2:]
							elif (dec==len(strline)-2):
								decText = strline[:-2]
							else:
								raise SystemExit,'-- at unexpected position in line '+str(self.lineNum)+' of '+self.moduleName+'.pyr. -- without an = is treated as dec, but must either be at the beginning or the end of the line.'
							when = decText.find(' when ')
							if (when>-1):
								varname = self.conversionMagic(decText[:when].strip(), startPointers=0, maxTokens=1)
								if varname.__class__!=str:
									varname=str(varname)
							
								if self.sgMore:
									self.scriptVer = 'sg'
								else:
									self.scriptVer = 'normal'
								condition = self.conversionMagic(incText[when+6:].strip(), startPointers=1)
								self.scriptVer='sg'
								if condition.__class__!=str:
									condition=str(condition)
								if (self.sgMore):
									if (condition.lower() == "true"):
										condition="1";
									elif (condition.lower() == "false"):
										condition="0";
									sss = sss = varname+" "+condition+" mult"
								else:
									if (condition.lower() == "1"):
										condition="true";
									elif (condition.lower() == "0"):
										condition="false";
									elif ((condition.lower()!="true") and (condition.lower()!="false")):
										if (not condition.endswith(self.conditionEnders)):
											print self.conditionWarningMessage(strline[when+6:].strip(), condition)
									if (self.inConditions>0):
										sss = varname+" dupbool "+condition+" and" #removed " mult" 5/8/2008
									else:
										sss = varname+" "+condition #removed " mult" 5/8/2008
								
								if (not self.sgMore):
									afterStore = " dropbool"
								
							else:
								sss=self.conversionMagic(decText, startPointers=0, maxTokens=1)	#maxTokens should cause an error to be thrown if more than one token is found.
							
							if (sss==None):
								raise SystemExit,'Error: Nothing recognized after ++ on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
							if sss.__class__!=str:
								sss=str(sss)
							backn = '\n'
							if self.combineLines:
								backn=' '
							self.scp_write(beforeLine+sss+afterAddress+" dec"+afterStore+backn)
						else:
							#error, you're leaving something on the stack?
							raise SystemExit,'Error: You\'re leaving something on the stack in line '+str(self.lineNum)+' of '+self.moduleName+'.pyr. Note: At the moment PyBot does not understand carrying stack values over to another operation on a separate line, so, this is currently considered an error by PyBot.'
						
		while numTabs<self.blockTab:
			self.lineTabs=self.blockTab-1
			st = self.blocks.pop()
			if st.__class__==str:
				self.scp_write(st)
			else:
				if st==1:		#elif - pop three
					output = 'dropbool dropbool dropbool\n' #'add add dup ^ inc\n'
					self.inConditions-=3 #bugfix: was 2
					self.branchValueSet=0
				elif st==3:
					output='dropbool dropbool\n' #'add dup ^ inc\n'
					self.inConditions-=2
					self.branchValueSet=0
				else:	#if or else - pop one
					output= 'dropbool\n' #'dup ^ inc\n'
					self.inConditions-=1
					self.branchValueSet=0
				self.scp_write(output)
			st = self.extraTextBlocks.pop()
			if st[0]!=None:
				self.extraText.append(st[0])
			if st[1]==1:
				self.writeToExtraText=0
			self.blockTab-=1
	
	def replaceNotInString(self, strline, char1, char2):
		startSpacePos=0
		#print "replaceNotInString('"+strline+"','"+char1+"','"+char2+"')"
		
		while (1):
			spacepos = self.finds(strline, char1, startSpacePos)	#self.finds(strline,char1,startSpacePos)
			#print "startSpacePos is "+str(startSpacePos)+" strline is '"+strline+"' spacepos is "+str(spacepos)+" len(strline) is "+str(len(strline))+"."
			if (spacepos>-1 and spacepos<len(strline)-1):
				#print "Checking for char2."
				if (strline[spacepos+1]==char2):
					#print "Found char2, changing startSpacePos to "+str(startSpacePos)+"."
					strline=strline[0:spacepos]+strline[spacepos+1:]
					startSpacePos = spacepos+1
				else:
					#print "Found nothing, changing startSpacePos to "+str(startSpacePos)+"."
					startSpacePos=spacepos+1
			else:
				break
		return strline
	
	#returns true if whatever's at 'pos' is inside a string. This is used so we can correctly detect assignment,
	#as opposed to someone saying sysmessage("Your color="+color). *wink*
	def inquote(self, arg, pos):
		retval=0
		start=0
		stringstart=None
		stringstartchar=None
		for a in range(0,len(arg)):
			if arg[a]=='"' or arg[a]=="'":
				if stringstart==None:
					stringstart=a
					stringend=None
					stringstartchar=arg[a]
				elif arg[a]==stringstartchar:
					stringstart=None
			if pos==a:
				if stringstart==None:
					return 0
				else:
					return 1
		return 0
	
	def findn(self, arg, findchar, startat):
		retval=-1
		if len(arg)<=startat:
			return -1
		recursechar=arg[startat]
		stringstart=None
		stringstartchar=None
		recurses=0
		for a in range(startat+1,len(arg)):
			if arg[a]==findchar:
				if stringstart==None:
					if recurses==0:
						return a
					else:
						recurses-=1
			elif arg[a]==recursechar:
				if stringstart==None:
					recurses+=1
			elif arg[a]=='"' or arg[a]=="'":
				if stringstart==None:
					stringstart=a
					stringend=None
					stringstartchar=arg[a]
				elif arg[a]==stringstartchar:
					if arg[a-1]=="\\":
						pass
					else:
						stringstart=None
		return -1
		
	def finds(self, arg, findchar, startat):
		retval=-1
		if len(arg)<=startat:
			return -1
		stringstart=None
		stringstartchar=None
		for a in range(startat,len(arg)):
			if arg[a]==findchar:
				if stringstart==None:
					return a
			elif arg[a]=='"' or arg[a]=="'":
				if stringstart==None:
					stringstart=a
					stringend=None
					stringstartchar=arg[a]
				elif arg[a]==stringstartchar:
					if arg[a-1]=="\\":
						pass
					else:
						stringstart=None
		return -1
	
	def replace(self, arg, frm, to):
		#print 'replacing '+frm+' with '+to+' in '+arg+'.'
		retval=arg
		comment=retval.find('\'')
		mins=0
		#a foobar flies,foobar,arf		- frmpos 15, frminto -1, topos -1
		#a foobar flies,foobar,foobart	- frmpos 15, frminto 0, topos -1
		#a foobar flies,foobar,			- frmpos 15, frminto -1, topos ?
		frmpos = retval.lower().find(frm,mins)
		frminto = to.lower().find(frm)
		topos = retval.lower().find(to,frmpos-frminto)
		#if (frm=="self." and frmpos>-1):
		#	print "arg is ("+arg+") frm is ("+frm+") to is ("+to+") frmpos is ("+str(frmpos)+") frminto is ("+str(frminto)+") topos is ("+str(topos)+") comment is ("+str(comment)+")"
		while topos+frminto==frmpos and frmpos>-1 and topos>-1 and len(to)>0:
			mins=frmpos+len(frm)
			frmpos = retval.lower().find(frm,mins)
			frminto = to.lower().find(frm)
			topos = retval.lower().find(to,frmpos-frminto)
			comment=retval.find('\'')
			#if (frm=="self."):
			#	print "in first while now: arg is ("+arg+") frm is ("+frm+") to is ("+to+") frmpos is ("+str(frmpos)+") frminto is ("+str(frminto)+") topos is ("+str(topos)+") comment is ("+str(comment)+") mins is ("+str(mins)+")"
		while frmpos>-1 and (comment==-1 or comment>frmpos):
			if not self.inquote(retval,frmpos):
				retval = retval[0:frmpos]+to+retval[frmpos+len(frm):]
				mins=frmpos+len(to)
				#if (frm=="self."):
				#	print "replaced. Now arg is ("+arg+") frm is ("+frm+") to is ("+to+") frmpos is ("+str(frmpos)+") frminto is ("+str(frminto)+") topos is ("+str(topos)+") comment is ("+str(comment)+") mins is ("+str(mins)+")"
			#else:
				#if (frm=="self."):
				#	print "is inquote."
			mins+=1
			frmpos = retval.lower().find(frm,mins)
			comment=retval.find('\'')
		return retval
		
	#This is a custom split function, which splits by commas, but recognizes strings and ignores commas in them.
	def split(self, arg):
		recurses=0
		retval=[]
		start=0
		stringstart=None
		stringstartchar=None
		for a in range(0,len(arg)):
			if arg[a]==')':
				if stringstart==None:
					if recurses>0:
						recurses-=1
					else:
						lastarg=arg[start:a]
						if len(lastarg)>0:
							retval.append(lastarg)
						return (retval, a+1)
			elif arg[a]=='(':
				if stringstart==None:
					recurses+=1
			elif arg[a]==',':
				if stringstart==None and recurses==0:
					retval.append(arg[start:a])
					start=a+1
			elif arg[a]=='"' or arg[a]=="'":
				if stringstart==None:
					stringstart=a
					stringend=None
					stringstartchar=arg[a]
				elif arg[a]==stringstartchar:
					if arg[a-1]=="\\":
						pass
					else:
						stringstart=None
		lastarg=arg[start:]
		if len(lastarg)>0:
			retval.append(lastarg)
		return (retval, len(arg)+1)
	
	#This counts the number of tabs preceding the code on a line.
	def countTabs(self, arg):
		numTabs=0
		for a in arg:
			if a=='\t':
				numTabs+=1
			else:
				return numTabs
		return numTabs
	
	#Auto-return and/or recursed should cause it to return a tuple containing the end pos.
	def conversionMagic(self, inputString, startPointers=1, autoReturn=0, maxTokens=-1, recursed=0):
		if self.debug:
			print 'conversionMagic("'+inputString+'", startPointers='+str(startPointers)+', autoReturn='+str(autoReturn)+', maxTokens='+str(maxTokens)+', recursed='+str(recursed)+') on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
		#maxTokens=0 disables maxTokens. 1 makes it throw an error if there are any tokens after the first.
		#autoReturn=0 disables autoReturn. 1 makes it return the first token that is stored.
		#Numerical tokens should be converted to strings after, not before, being returned. Complex tokens are fundamentally strings too - just returned from a subcall of this function.
		#This should throw an error if there are invalid token combinations, like two non-operator tokens in a row, or 'foo + * bar'.
		#Functions (not unary operators) call a separate method to split their args. That function will call this again for each comma-separated part of the function call parameters.
		#There is special case handling to determine if - is being used as a unary or binary operator. This is simple enough. If there's a non-operator token preceding the -, it is binary. If not, is is unary.
		#There is also a special case handling to make unary - combine with a numerical token instead of taking its normal route: e.g. "- 5" should become "-5" instead of "5 -".
		
		#See AnimalMinimalis.pyr for details of how this should operate.
		
		#This should always return either a string or a number. If it returns a number, the caller can detect it with if word.__class__!=str:. This can be used to implement the special case handling for unary - mentioned above, and will need to be done to convert the numbers to strings for token combining.
		
		prevToken = None
		opToken = None
		thisToken = None
		thisTokenIsNumber = None
		opType = None
		opStandard = None
		opTranslation = None
		
		retstr = ''
		varstart=None
		varend=None
		varhassymbols=0
		hexstart=None
		decstart=None
		pointers=startPointers
		pointerParenAdjust = 0
		
		specialOperatorChars = '/*~+-><=!%'
		
		#while len(inputString)>=1 and inputString[0]=='\t':
			#inputString=inputString[1:]
		inputString = inputString.rstrip('\t \n')
		startCharsRemoved = len(inputString)
		inputString = inputString.lstrip('\t \n')
		startCharsRemoved -= len(inputString)
		inputString = inputString+"\n"		# This elegantly causes tokens ending at the end of the line to be recognized without needing to duplicate code or add special checks (\n is just recognized as whitespace).
		skipToA = -1
		for a in range(0,len(inputString)):
			if skipToA>a:
				continue
			newToken = None
			#	Identify any newly completed token. Determine what it is.
			if varstart!=None and varend==None and (varhassymbols==1 or (string.letters.find(inputString[a])==-1 and inputString[a]!='_' and inputString[a]!='$' and inputString[a]!='&' and string.digits.find(inputString[a])==-1)) and (varhassymbols==0 or specialOperatorChars.find(inputString[a])==-1):
				newToken = inputString[varstart:a].strip()
				thisTokenIsNumber=None
				varstart=None
				varend=a
				#What is this token?
				if (newToken=="-"):	#special case
					if self.debug:
						print 'NewToken recognized as -. opToken is "'+str(opToken)+'".'
					if thisToken==None:
						newToken="-u"
					else:
						newToken="-b"
				myOpType = None
				myOpStandard = None
				myOpTranslation = None
				try:
					myOpType = self.patterns[self.scriptVer][newToken]['optype']
					myOpStandard = self.patterns[self.scriptVer][newToken]['standard']
					myOpTranslation = self.patterns[self.scriptVer][newToken]['translation']
				except KeyError:
					pass
				if (myOpType!=None or myOpType==self.optype_direct):	#operator
					if (myOpStandard==0):
						raise SystemExit,'Non-standard operators are not yet supported ("'+newToken+'"). This error is in ('+inputString+'), on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
					#TODO: Should the pointers variable be checked on an operator to throw errors, or would it naturally have different values?
					if (myOpType==self.optype_unary):
						if (opToken==None and thisToken!=None):
							#error condition
							raise SystemExit,'Unary operators ('+newToken+') must follow either nothing or another operator (not '+thisToken+'). This error is in ('+inputString+'), on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						else:
							#Recurse instead of overwriting opToken.
							if self.debug:
								print "Identified unary operator. Recursing on it."
							retTuple = None
							if (inputString[a]=='('):
								#~ prevToken = None
								#~ opToken = newToken
								#~ opType = myOpType
								#~ opTranslation = myOpTranslation
								#~ thisToken = None
								#~ newToken = None
								#~ thisTokenIsNumber=None
								retTuple = self.conversionMagic(inputString[a+1:], pointers, recursed=1)
								if self.debug:
									print 'Call returned '+str(retTuple)+'.'
								skipToA = a + retTuple[1] + 1
								if self.debug:
									print 'Skipping from position '+str(a)+' to '+str(skipToA)+' in "'+inputString+'".'
								if retTuple[0].__class__!=str:	#A number
									if newToken=="-u":
										newToken="-"+str(retTuple[0])
									else:
										newToken=str(retTuple[0])+" "+myOpTranslation
								else:
									newToken=retTuple[0]+" "+myOpTranslation
							else:
								#retTuple = self.conversionMagic(inputString[a+1:], pointers, recursed=1)
								retTuple = self.conversionMagic(inputString[a:], pointers, autoReturn=1)
								if self.debug:
									print 'Call returned '+str(retTuple)+'.'
								skipToA = a + retTuple[1]
								if self.debug:
									print 'Skipping from position '+str(a)+' to '+str(skipToA)+' in "'+inputString+'".'
								if retTuple[0].__class__!=str:	#A number
									if newToken=="-u":
										newToken="-"+str(retTuple[0])
									else:
										newToken=str(retTuple[0])+" "+myOpTranslation
								else:
									newToken=retTuple[0]+" "+myOpTranslation
					elif (myOpType==self.optype_binary):
						if (opToken==None):
							if thisToken!=None:
								if self.debug:
									print 'Recording binary operator '+newToken+', prev token "'+thisToken+'".'
								prevToken = thisToken
								opToken = newToken
								opType = myOpType
								opTranslation = myOpTranslation
								thisToken = None
								newToken = None
								thisTokenIsNumber=None
							else:
								raise SystemExit,'Binary operators (such as '+newToken+') must follow another non-operator token (there was none). This error is in ('+inputString+'), on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						else:
							raise SystemExit,'Binary operators (such as '+str(newToken)+') must not follow another operator (which was '+str(opToken)+'). This error is in ('+inputString+'), on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
					elif (myOpType>0):	#function
						if self.debug:
							print 'Parsing function call to '+newToken+', prev token "'+str(thisToken)+'".'
						retTuple = self.parseArgs(myOpTranslation, myOpType, inputString[a+1:], pointers)
						skipToA = a + retTuple[1] + 1
						if self.debug:
							print 'Skipping from position '+str(a)+' to '+str(skipToA)+' ('+str(a)+'+'+str(retTuple[1])+'+1) in "'+inputString+'".'
						if retTuple[0].__class__!=str:	#A number
							newToken = str(retTuple[0])
						else:
							newToken = retTuple[0]
						thisTokenIsNumber=None
						
				else:
					#variable name
					isdefined = self.getDefined(pointers, newToken)
					if isdefined!=newToken:
						newToken = isdefined
						thisTokenIsNumber=None
					else:
						if pointers<-1:
							raise SystemExit,'Trying to get the address of the address of a variable is illegal. This error is in ('+inputString+'), on line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						if (opType==None):
							if (thisToken!=None):
								#non-op token after another non-op token: illegal
								raise SystemExit,'You cannot have a non-operator token (such as '+newToken+') following another non-operator token (such as '+thisToken+'). This error is in ('+inputString+'), on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
							else:
								if (myOpType==self.optype_direct):
									newToken=myOpTranslation
								else:
									newToken = self.translatePointers(pointers, "."+newToken)
								thisTokenIsNumber=None
								if self.debug:
									print 'Identified variable-name token "'+newToken+'".'
							
						elif (opType==self.optype_binary):
							if (myOpType==self.optype_direct):
								newToken=myOpTranslation
							else:
								newToken = self.translatePointers(pointers, "."+newToken)
							thisTokenIsNumber=None
							if self.debug:
								print '(Have binary operator) Identified variable-name token "'+newToken+'".'
						elif (opType==self.optype_unary):
							raise SystemExit,'Sanity check failure: Unary operator recorded as operator without recursion eliminating checks on the next variable? Triggered by line ('+inputString+'), on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
						else:
							raise SystemExit,'Sanity check failure: Function operator recorded as operator without recursion eliminating checks on the next variable? Triggered by line ('+inputString+'), on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				pointers=startPointers	
			elif hexstart!=None and string.hexdigits.find(inputString[a].upper())==-1:
				#end token
				value = int(inputString[hexstart:a], 16)
				newToken = str(value)
				pointers-=1	#because it's a number
				if pointers<0:
					#raise SystemExit,'Trying to get the address of a number is illegal. This error is in ('+inputString+'), on line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
					pointers=0
				if pointers!=0:
					newToken=self.translatePointers(pointers, newToken)
				else:
					thisTokenIsNumber = value
				pointers=startPointers
				hexstart=None
				if self.debug:
					print 'Identified hex number token "'+newToken+'".'						
			elif decstart!=None and string.digits.find(inputString[a])==-1:
				#end token
				value = int(inputString[decstart:a])
				newToken = str(value)
				pointers-=1	#because it's a number
				if pointers<0:
					#raise SystemExit,'Trying to get the address of a number is illegal. This error is in ('+inputString+'), on line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
					pointers=0
				if pointers!=0:
					newToken=self.translatePointers(pointers, newToken)
				else:
					thisTokenIsNumber = value
				decstart=None
				pointers=startPointers
				if self.debug:
					print 'Identified decimal number token "'+newToken+'".'
			if (newToken!=None):
				#	If we have an operator token already, process the tokens and store the result.
				if (thisToken!=None):
					raise SystemExit,'Can\'t have multiple variable or number tokens after each other without operators. This error is in ('+inputString+'), on line: '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				thisToken = newToken
				if self.debug:
					print 'Stored token "'+newToken+'".'
				if (opToken!=None):
					if (opType==self.optype_binary):
						thisToken = prevToken+" "+thisToken+" "+opTranslation
						thisTokenIsNumber = None
						if self.debug:
							print 'Processed binary operation, which became token "'+thisToken+'" (from prevToken "'+prevToken+'", opTranslation "'+opTranslation+'", thisToken "'+newToken+'").'
						prevToken = None
						opToken = None
						
				if (autoReturn>0):
					autoReturn-=1
					if (autoReturn<=0):
						if (skipToA > a):
							a = skipToA
						if self.debug:
							print 'Auto-returning "'+thisToken+'", '+str(a)+'+1+'+str(startCharsRemoved)
						if (thisTokenIsNumber!=None):
							return (thisTokenIsNumber, a+startCharsRemoved)
						else:
							return (thisToken, a+startCharsRemoved)
				elif (maxTokens>-1):
					maxTokens-=1
					if (maxTokens<0):
						raise SystemExit,'Only one token expected in "'+inputString+'" on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				pointers=startPointers
			if (inputString[a]!='&' and inputString[a]!='$' and inputString[a]!='('):
				pointerParenAdjust = 0
			
			if varstart!=None or skipToA>a:
				pass
			elif inputString[a]=='0' and varstart==None:
				#hex num started here
				if decstart==None and hexstart==None:
					hexstart=a
			elif string.digits.find(inputString[a])>-1:
				#numbers
				if hexstart==None and decstart==None:
					decstart=a
			elif hexstart!=None and string.hexdigits.find(inputString[a].upper())>-1:
				pass	#continue
			elif inputString[a]=='&':
				pointers-=1
				pointerParenAdjust-=1
			elif inputString[a]=='$':
				pointers+=1
				pointerParenAdjust+=1
			elif (string.letters.find(inputString[a])>-1):
				if varstart==None:
					if self.debug:
						print 'Variable name or operator starting at '+str(a)+'.'
					varstart=a
					varend=None
					varhassymbols = 0
			elif (specialOperatorChars.find(inputString[a])>-1):
				if varstart==None:
					if self.debug:
						print 'Operator starting at '+str(a)+'.'
					varstart=a
					varend=None
					varhassymbols = 1
			elif inputString[a]=='(':
				if self.debug:
					print 'Recursing due to paren.'						
				retTuple = self.conversionMagic(inputString[a+1:], startPointers, recursed=1)
				if self.debug:
					print 'Call returned '+str(retTuple)+'.'
				skipToA = a + retTuple[1] + 1
				if self.debug:
					print 'Going to skip to '+str(skipToA)+' from '+str(a)+' in string "'+inputString+'"'
				if retTuple[0].__class__!=str:	#A number
					thisToken = str(retTuple[0])
					thisTokenIsNumber = retTuple[0]
				else:
					thisToken = retTuple[0]
					thisTokenIsNumber = None
				if (pointerParenAdjust>0):
					while (pointerParenAdjust>0):
						thisToken += " * "
						pointerParenAdjust-=1
				elif (pointerParenAdjust<0):
					raise SystemExit,'Can\'t get the address of the contents of parens. You can, however, use the contents of parens as a pointer by prefixing it with a $. Error is on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
				if prevToken!=None:
					if self.debug:
						print 'PrevToken is now "'+prevToken+'", and thisToken is "'+thisToken+'".'
				else:
					if self.debug:
						print 'ThisToken is now "'+thisToken+'".'
				
				if (opToken!=None):
					if (opType==self.optype_binary):
						if thisToken!=None:
							if (prevToken!=None):
								tempToken = thisToken
								thisToken = prevToken+" "+thisToken+" "+opTranslation
								thisTokenIsNumber = None
								if self.debug:
									print 'After recursion, processed binary operation, which became token "'+thisToken+'" (from prevToken "'+prevToken+'", opTranslation "'+opTranslation+'", thisToken "'+tempToken+'").'
								prevToken = None
								opToken = None
							else:
								raise SystemExit,'Sanity check failure: prevToken is None when thisToken is "'+thisToken+'" and opToken is "'+opToken+'" on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
					elif (opType==self.optype_unary):
						tempToken = thisToken
						thisToken = thisToken+" "+opTranslation
						thisTokenIsNumber = None
						if self.debug:
							print 'After recursion, processed unary operation, which became token "'+thisToken+'" (from  opTranslation "'+opTranslation+'", thisToken "'+tempToken+'").'
						prevToken = None
						opToken = None
			elif inputString[a]==')':
				#return
				# 01234567890123
				#"(foo bar) zort"
				# will return 8
				if (recursed):
					if (thisToken==None):
						thisToken=""
					if self.debug:
						print 'End of recursion: Returning "'+str(thisToken)+'", '+str(a)+'+1+'+str(startCharsRemoved)
					if (thisTokenIsNumber!=None):
						return (thisTokenIsNumber, a+1+startCharsRemoved)
					else:
						return (thisToken, a+1+startCharsRemoved)
				else:
					raise SystemExit,'Unexpected ) on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
			elif inputString[a]==' ' or inputString[a]=='\t' or inputString[a]=='\n':
				pass #retstr+=inputString[a]	
			else:
				raise SystemExit,'Unexpected or unknown character '+inputString[a]+' on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
		if (recursed>0):
			raise SystemExit,'Missing ) in "'+inputString+'" on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
		if (autoReturn>0):
			autoReturn-=1
			if (autoReturn<=0):
				if self.debug:
					print 'Auto-returning (at end of string).'
				if (thisTokenIsNumber!=None):
					return (thisTokenIsNumber, len(inputString)+startCharsRemoved)
				else:
					return (thisToken, len(inputString)+startCharsRemoved)
		if self.debug:
			print 'End of string: Returning "'+str(thisToken)+'".'
		return thisToken
	
	def parseArgs(self, functionName, numExpectedParams, inputString, startPointers):
		if self.debug:
			print 'parseArgs(functionName="'+functionName+'", numExpectedParams='+str(numExpectedParams)+', inputString="'+inputString+'", startPointers='+str(startPointers)+')'
		pointers = startPointers
		paramsRead = 0
		startCharsRemoved = len(inputString)
		inputString = inputString.lstrip('\t \n')
		startCharsRemoved -= len(inputString)
		charsToSkip = startCharsRemoved
		retval = ""
		if (len(inputString)>0):
			tupleVal = self.split(inputString)
			params = tupleVal[0]
			charsToSkip += tupleVal[1]
			index=0
			for param in params:
				if self.debug:
					print 'Converting param "'+param+'"'
				value = self.conversionMagic(param, startPointers)
				if self.debug:
					print 'We got this: "'+str(value)+'"'
				if value.__class__!=str:
					value=str(value)
				retval += value+" "
				paramsRead+=1
				index+=1
			retval+=functionName
		if (paramsRead!=numExpectedParams):
			raise SystemExit,'Incorrect number of params to '+functionName+' (expected '+str(numExpectedParams)+', got '+str(paramsRead)+' on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
		if self.debug:
			print 'parseArgs returning "'+retval+'"'
		return (retval, charsToSkip)
	
	def translatePointers(self, pointers, token):
		if (pointers==0):
			if self.debug:
				print 'translatePointers('+str(pointers)+', "'+token+'") returning "'+token+'"'
			return token
		elif (pointers==1):
			if self.debug:
				print 'translatePointers('+str(pointers)+', "'+token+'") returning "*'+token+'"'
			return "*"+token
		elif (pointers>1):
			origToken = token
			token = "*"+token
			while (pointers>1):
				token = token+" *"
				pointers-=1
			if self.debug:
				print 'translatePointers('+str(pointers)+', "'+origToken+'") returning "'+token+'"'
			return token
		else:
			raise SystemExit,'Error: Attempt to get the address of an address or integer or the like, on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr.'
	
	def getDefined(self, pointers, name):
		try:
			value = self.defined[name]
			return self.conversionMagic(value[0], startPointers=pointers+value[1])
		except KeyError:
			return name
	
	def conditionWarningMessage(self, originalCondition, translatedCondition):
		return '\nWarning: The condition on line '+str(self.lineNum)+' of '+self.moduleName+'.pyr does not appear to end with a boolean result. What this means is: You have to write, for example, foo when (multi!=0), instead of foo when (multi).\nOriginal condition: '+originalCondition+'\nTranslated condition: '+translatedCondition+'\n\n'
	
def hasContent(line):
	for a in line:
		if a!=' ' and a!='\t':
			return 1
	return 0


#The first thing run! Calls main(). Lucky you, you don't need anything like this in your PyBot scripts. :P
if __name__ == '__main__': main()